import numpy as np
import torch
import torchvision
import torchvision.transforms as transforms
import torch.optim as optim

#https://pytorch.org/tutorials/recipes/recipes/tensorboard_with_pytorch.html
# https://pytorch.org/tutorials/intermediate/tensorboard_tutorial.html
### pip install tensorboard
from torch.utils.tensorboard import SummaryWriter
writer = SummaryWriter()
# Writer will output to ./runs/ directory by default.
# you can use e.g. writer = SummaryWriter('runs/experiment_1')

IMG_SIZE = 32
BATH_SIZE = 8

class_names = {
    0: "Speed limit (20km/h)",
    1: "Speed limit (30km/h)",
    2: "Speed limit (50km/h)",
    3: "Speed limit (60km/h)",
    4: "Speed limit (70km/h)",
    5: "Speed limit (80km/h)",
    6: "End of speed limit (80km/h)",
    7: "Speed limit (100km/h)",
    8: "Speed limit (120km/h)",
    9: "No passing",
    10: "No passing veh over 3.5 tons",
    11: "Right-of-way at intersection",
    12: "Priority road",
    13: "Yield",
    14: "Stop",
    15: "No vehicles",
    16: "Veh > 3.5 tons prohibited",
    17: "No entry",
    18: "General caution",
    19: "Dangerous curve left",
    20: "Dangerous curve right",
    21: "Double curve",
    22: "Bumpy road",
    23: "Slippery road",
    24: "Road narrows on the right",
    25: "Road work",
    26: "Traffic signals",
    27: "Pedestrians",
    28: "Children crossing",
    29: "Bicycles crossing",
    30: "Beware of ice/snow",
    31: "Wild animals crossing",
    32: "End speed + passing limits",
    33: "Turn right ahead",
    34: "Turn left ahead",
    35: "Ahead only",
    36: "Go straight or right",
    37: "Go straight or left",
    38: "Keep right",
    39: "Keep left",
    40: "Roundabout mandatory",
    41: "End of no passing",
    42: "End no passing veh > 3.5 tons"
}
transform = transforms.Compose(
    [transforms.ToTensor(),
     transforms.Resize((IMG_SIZE, IMG_SIZE), antialias=True),
     ])

trainset = torchvision.datasets.GTSRB(root='./data', split="train",
                                        download=True, transform=transform)

testset = torchvision.datasets.GTSRB(root='./data', split="test",
                                       download=True, transform=transform)


import matplotlib.pyplot as plt
import numpy as np

figure = plt.figure(figsize=(8, 8))
for i in range(1, 26):
    img, label = trainset[np.random.randint(0, len(trainset))]
    figure.add_subplot(5, 5, i)
    print(img.size())
    img_t = np.transpose(img, (1, 2, 0))
    print(type(img))
    #img_t = torch.squeeze(img, 1)
    print(img_t.size())
    plt.title(class_names.get(label, "Unknown class ID"))
    plt.axis("off")
    plt.imshow(img_t, cmap="gray")    
    print(i)
plt.show()

testloader = torch.utils.data.DataLoader(testset, batch_size=BATH_SIZE,
                                         shuffle=False)

trainloader = torch.utils.data.DataLoader(trainset, batch_size=BATH_SIZE,
                                          shuffle=True)

examples = iter(trainloader)
examples_grid_data, examples_grid_label = next(examples)
t_grid = torchvision.utils.make_grid(examples_grid_data)
writer.add_image("GTSRB", t_grid)

import torch.nn as nn
import torch.nn.functional as F


class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 43)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1) # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

net = Net()
       
writer.add_graph(net, examples_grid_data)
 
criterion = torch.nn.CrossEntropyLoss()
optimizer = optim.SGD(net.parameters(), lr=0.01)
total_step = 0
for epoch in range(4):  # loop over the dataset multiple times
    
    running_loss = 0.0
    for i, data in enumerate(trainloader, 0):
        # get the inputs; data is a list of [inputs, labels]
        inputs, labels = data

        # zero the parameter gradients
        optimizer.zero_grad()

        # forward + backward + optimize
        outputs = net(inputs)
        loss = criterion(outputs, labels)
        loss.backward()
        optimizer.step()

        # print statistics
        running_loss += loss.item()       
        if i % 20 == 19:    # print every 20 mini-batches
            writer.add_scalar("Loss/train", running_loss/20, epoch * len(trainloader) + i)            
            print(f'[{epoch + 1}, {i + 1:5d}] loss: {running_loss / 20:.3f}')
            running_loss = 0.0
            
print('Finished Training')  
writer.flush()
writer.close()

correct = 0
total = 0
# since we're not training, we don't need to calculate the gradients for our outputs
with torch.no_grad():
    for data in testloader:
        images, labels = data
        # calculate outputs by running images through the network
        outputs = net(images)
        # the class with the highest energy is what we choose as prediction
        _, predicted = torch.max(outputs.data, 1)
        total += labels.size(0)
        correct += (predicted == labels).sum().item()

print(f'Accuracy of the network on the 10000 test images: {100 * correct // total} %')


